/*    %W% %G%
================================================================================
  Description :
  Author      : Ryu, Byeong-Soon
================================================================================
*/

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include "socket.h"
#include "channel.h"

/*#define _NO_FILTER_ */

/*#define TRACE_ON*/
#include "debug.h"

/* number of additional requests : for hangul font */
#define ADDITIONAL_REQUESTS		2

static Channel  *_Channels;
static int  _NumChannels = 0;

int ByteOrder;

int  VXSockFD;

FILE *log;

short
mem_short(char *ptr, int offset)
{
	short  *sptr;

	sptr = (short *) (ptr+offset);
	return  *sptr;
}

long
mem_long(char *ptr, int offset)
{
	long  *lptr;

	lptr = (long *) (ptr+offset);

	return *lptr;
}

void
mem_set_short(char *ptr, int offset, short val)
{
	if (ByteOrder == OP_LSB_FIRST)
	{
		ptr[offset] = (val & 0x00ff);
		ptr[offset+1]   = (val & 0xff00) >> 8;
	}
	else
	{
		ptr[offset]   = (val & 0xff00) >> 8;
		ptr[offset+1] = (val & 0x00ff);
	}
}

void
mem_set_long(char *ptr, int offset, long val)
{
	if (ByteOrder == OP_LSB_FIRST)
	{
		ptr[offset] = (val & 0x000000ff);
		ptr[offset+1] = (val & 0x0000ff00) >> 8;
		ptr[offset+2] = (val & 0x00ff0000) >> 16;
		ptr[offset+3]   = (val & 0xff000000) >> 24;
	}
	else
	{
		ptr[offset]   = (val & 0xff000000) >> 24;
		ptr[offset+1] = (val & 0x00ff0000) >> 16;
		ptr[offset+2] = (val & 0x0000ff00) >> 8;
		ptr[offset+3] = (val & 0x000000ff);
	}
}

void
VXInitChannel()
{
	_Channels = NULL;
	_NumChannels = 0;
	/*
	log = fopen("log", "wb");
	*/
}

void
VXNewChannel()
{
	int  port, cfd, sfd;
	char  addr[100];
	Channel  *new_cl;
	int  len;
	struct sockaddr_in  from;

	TRACE("new channel\n");
	len = sizeof(from);
	getsockname(VXSockFD, (struct sockaddr *) &from, &len);

	cfd = accept(VXSockFD, (struct sockaddr *) &from, &len);
	if(cfd < 0) {
		perror("accept");
		return;
	}

	VXGetAddress(addr, &port);
	TRACE2("connect to X-server: %s, %d\n", addr, port);
	sfd = SockInitializeClient(addr, port);
	if(sfd < 0) {
		return;
	}

	new_cl = (Channel *) malloc(sizeof(Channel));
	new_cl->next = _Channels;
	_Channels = new_cl;
	_NumChannels++;

	new_cl->cfd = cfd;
	new_cl->sfd = sfd;

	/* clear new channel */
	new_cl->c_connected = 0;
	new_cl->s_connected = 0;
	new_cl->ci_buff_size = 0;
	new_cl->co_buff_size = 0;
	new_cl->si_buff_size = 0;
	new_cl->so_buff_size = 0;
	new_cl->c_req_len = 0;
	new_cl->s_req_len = 0;
	new_cl->ctemp = NULL;
	new_cl->c_str_req_len = 0;
	new_cl->c_str_len = 0;
	new_cl->c_ipos = 0;
	new_cl->c_opos = 0;
	new_cl->s_ipos = 0;
	new_cl->s_opos = 0;
	new_cl->c_len_field = -1;
	new_cl->hangul = 0;
	new_cl->input_temp = 0;
	new_cl->font_loaded = 0;
	new_cl->font_gc = NULL;
}

void
VXFreeChannel(Channel *pre, Channel *ptr)
{
	TRACE("close channel\n");
	if(pre == NULL) {
		_Channels = ptr->next;
	} else {
		pre->next = ptr->next;
	}

	close(ptr->cfd);
	close(ptr->sfd);
	free(ptr);
	_NumChannels--;
}

void
hupsig()
{
	char  cmd[30];
	Channel  *ptr;

	for(ptr = _Channels; ptr != NULL; ptr = ptr->next) {
		close(ptr->cfd);
		close(ptr->sfd);
	}
	close(VXSockFD);

	/*
	sprintf(cmd, "/bin/rm -f %s", VXPortFile);
	system(cmd);
	*/

	exit(0);
}

void
VXMainLoop()
{
	int  mask, max_desc = ulimit(4, 0);
	fd_set  in_set, out_set, exc_set;
	Channel  *ptr;

	TRACE1("max_desc : %d\n", max_desc);

	signal(SIGHUP, hupsig);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, hupsig);
	signal(SIGTERM, hupsig);

	TRACE("main loop\n");
	while(1) {
		FD_ZERO(&in_set);
		FD_ZERO(&out_set);
		FD_ZERO(&exc_set);

		FD_SET(VXSockFD, &in_set);
		for(ptr = _Channels; ptr != NULL; ptr = ptr->next) {
			/* exceptions */
			FD_SET(ptr->sfd, &exc_set);
			FD_SET(ptr->cfd, &exc_set);

			/* outputs */
			if( ptr->so_buff_size > 0 ||
				ptr->ci_buff_size > 0 ||
				ptr->input_temp == 1
			) {
				FD_SET(ptr->sfd, &out_set);
			}
			if(ptr->co_buff_size > 0 || ptr->si_buff_size > 0)
				FD_SET(ptr->cfd, &out_set);

			/* inputs */
			if(ptr->si_buff_size < MAX_BUFFER)
				FD_SET(ptr->sfd, &in_set);
			if(ptr->ci_buff_size < MAX_BUFFER)
				FD_SET(ptr->cfd, &in_set);
		}

		if(select(max_desc, &in_set, &out_set, &exc_set, NULL) < 0) {
			perror("select");
			return;
		}

		VXProcessInput(&in_set, &out_set, &exc_set);
	}
}

void
VXProcessInput(fd_set *in_set, fd_set *out_set, fd_set *exc_set)
{
	Channel  *ptr, *old_ptr, *next;

	/* New connection? */
	if(FD_ISSET(VXSockFD, in_set)) {
		VXNewChannel();
	}

	/* Kick out the freaky folks */
	old_ptr = NULL;
	for(ptr = _Channels; ptr != NULL; ptr = next) {
		next = ptr->next;
		if(FD_ISSET(ptr->cfd, exc_set) || FD_ISSET(ptr->sfd, exc_set)) {
			FD_CLR(ptr->cfd, in_set);
			FD_CLR(ptr->cfd, out_set);
			FD_CLR(ptr->sfd, in_set);
			FD_CLR(ptr->sfd, out_set);
			VXFreeChannel(old_ptr, ptr);
		} else {
			old_ptr = ptr;
		}
	}

	old_ptr = NULL;
	for(ptr = _Channels; ptr != NULL; ptr = next) {
		next = ptr->next;

		if(FD_ISSET(ptr->sfd, in_set)) {
			if(VXInputServer(ptr) <= 0) {
				VXFreeChannel(old_ptr, ptr);
				ptr = old_ptr;
				continue;
			}
		}
		if(FD_ISSET(ptr->cfd, out_set)) {
			if(VXOutputClient(ptr) < 0) {
				VXFreeChannel(old_ptr, ptr);
				continue;
			}
		}

		if(FD_ISSET(ptr->cfd, in_set)) {
			if(VXInputClient(ptr) <= 0) {
				VXFreeChannel(old_ptr, ptr);
				ptr = old_ptr;
				continue;
			}
		}
		if(FD_ISSET(ptr->sfd, out_set)) {
			if(VXOutputServer(ptr) < 0) {
				VXFreeChannel(old_ptr, ptr);
				continue;
			}
		}
		old_ptr = ptr;
	}
}

int
VXInputClient(Channel *ptr)
{
	int  nleft, nread;
	int i;
	char *tptr;

	nleft = MAX_BUFFER - ptr->ci_buff_size;
	if(nleft == 0)
		return 0;

	nread = read(ptr->cfd, ptr->ci_buffer+ptr->ci_buff_size, nleft);
	if(nread < 0)
		return nread;
	TRACE2("input client: nread = %d nleft = %d\n", nread, nleft);
	/*
	if(nread > 1024)
		printf("IC: nread = %d\n", nread);
	fprintf(log, "**IC**");
	fwrite(ptr->ci_buffer+ptr->ci_buff_size, 1, nread, log);
	fflush(log);
	*/

	ptr->ci_buff_size += nread;

	return nread;
}

int
VXInputServer(Channel *ptr)
{
	int  nleft, nread;

	nleft = MAX_BUFFER - ptr->si_buff_size;
	if(nleft == 0)
		return 0;

	nread = read(ptr->sfd, ptr->si_buffer+ptr->si_buff_size, nleft);
	if(nread < 0)
		return nread;
	TRACE1("input server: nread = %d\n", nread);
	/*
	if(nread > 1024)
		printf("IS: nread = %d\n", nread);
	fprintf(log, "**IS**");
	fwrite(ptr->si_buffer+ptr->si_buff_size, 1, nread, log);
	fflush(log);
	*/

	ptr->si_buff_size += nread;

	return nread;
}

int
VXOutputClient(Channel *ptr)
{
	int  i, nwritten;

	VXFilterServerInput(ptr);
	if(ptr->co_buff_size == 0)
		return 0;

	nwritten = write(ptr->cfd, ptr->co_buffer, ptr->co_buff_size);
	if(nwritten < 0)
		return nwritten;
	TRACE2("output client: size = %d, nwritten = %d\n",
		ptr->co_buff_size, nwritten
	);
	/*
	if(nwritten > 1024)
		printf("OC: nwritten = %d\n", nwritten);
	fprintf(log, "**OC**");
	fwrite(ptr->co_buffer, 1, nwritten, log);
	fflush(log);
	*/

	ptr->co_buff_size -= nwritten;
	if(ptr->co_buff_size > 0)
		memmove(ptr->co_buffer, ptr->co_buffer+nwritten, ptr->co_buff_size);
	return  nwritten;
}

int
VXOutputServer(Channel *ptr)
{
	int  i, nwritten;

	VXFilterClientInput(ptr);
	if(ptr->so_buff_size == 0)
		return 0;

	nwritten = write(ptr->sfd, ptr->so_buffer, ptr->so_buff_size);
	if(nwritten < 0)
		return nwritten;
	/*
	if(nwritten > 1024)
		printf("OS: nwritten = %d\n", nwritten);
	TRACE2("output server: size = %d, nwritten = %d\n",
		ptr->so_buff_size, nwritten
	);
	fprintf(log, "**OS**");
	fwrite(ptr->so_buffer, 1, nwritten, log);
	fflush(log);
	*/

	ptr->so_buff_size -= nwritten;
	if(ptr->so_buff_size > 0)
		memmove(ptr->so_buffer, ptr->so_buffer+nwritten, ptr->so_buff_size);
	return  nwritten;
}

void
VXMemCopyClient(Channel *ptr, int len)
{
	memmove(ptr->so_buffer+ptr->so_buff_size, ptr->ci_buffer, len);
	memmove(ptr->ci_buffer, ptr->ci_buffer+len, ptr->ci_buff_size-len);
	ptr->ci_buff_size -= len;
	ptr->so_buff_size += len;
	ptr->c_req_len -= len;
	ptr->c_ipos += len;
}

void
VXMemCopyServer(Channel *ptr, int len)
{
	memmove(ptr->co_buffer+ptr->co_buff_size, ptr->si_buffer, len);
	memmove(ptr->si_buffer, ptr->si_buffer+len, ptr->si_buff_size-len);
	ptr->si_buff_size -= len;
	ptr->co_buff_size += len;
	ptr->s_req_len -= len;
	ptr->s_ipos += len;
}

void
VXFilterClientInput(Channel *ptr)
{
	int  len, dlen;

#ifndef	_NO_FILTER_
	if(ptr->input_temp) {
		TRACE("input_temp\n");
		/* len = min(available space, number of data to be output) */
		len = MAX_BUFFER - ptr->so_buff_size;
		if(ptr->c_opos < len)
			len = ptr->c_opos;
		if(len == 0)		/* buffer is full || no data */
			return;

		memmove(ptr->so_buffer+ptr->so_buff_size, ptr->ctemp, len);
		memmove(ptr->ctemp, ptr->ctemp+len, ptr->c_opos-len);
		ptr->c_opos -= len;
		ptr->so_buff_size += len;
		if(ptr->c_opos == 0) {
			free(ptr->ctemp);
			ptr->input_temp = 0;
			ptr->c_req_len = 0;
		}
		return;
	}

	/* len = min(available space, number of data to be output) */
	len = MAX_BUFFER - ptr->so_buff_size;
	if(ptr->ci_buff_size < len)
		len = ptr->ci_buff_size;
	if(len == 0)		/* buffer is full || no data */
		return;

	while(len > 0) {
		if(ptr->c_req_len == 0) {		/* new request */
			if(ptr->s_connected == 1 && ptr->font_loaded == 0) {
				VXInitHangulFont(ptr);
				ptr->font_loaded = 1;
			}
			ptr->c_ipos = 0;
			ptr->c_str_len = 0;
			ptr->opcode = ptr->ci_buffer[0] & 0x0000ff;
			if(ptr->c_connected == 0) {
				if(ptr->opcode == OP_MSB_FIRST || ptr->opcode == OP_LSB_FIRST) {
					ByteOrder = ptr->opcode;
					if (ByteOrder == OP_MSB_FIRST)
						fprintf(stderr, "ByteOrder = MSB_FIRST\n");
					else
						fprintf(stderr, "ByteOrder = LSB_FIRST\n");
					ptr->byte_order = ptr->opcode;

					/* length field is not on memory */
					if(ptr->ci_buff_size < 10) {
						TRACE("*** NO LENGTH FIELD ***\n");
						break;
					}

					dlen = mem_short(ptr->ci_buffer, 6);	/* len auth-name */
					dlen += mem_short(ptr->ci_buffer, 8);	/* len auth-data */
					ptr->c_req_len = dlen * 4 + 12;
				} else {	/* pass error, as it is, to the server */
					VXMemCopyClient(ptr, len);
					len = 0;
				}
			} else {
				/* length field is not on memory */
				if(ptr->ci_buff_size < 4) {
					TRACE("*** NO LENGTH FIELD ***\n");
					break;
				}

				dlen = mem_short(ptr->ci_buffer, 2);
				ptr->c_req_len = dlen * 4;
			}
			TRACE2("OP_CODE = %u, len = %d\n", (0x000000ff & ptr->opcode), ptr->c_req_len);
		} else {	/* body of request */
			switch(ptr->opcode) {
			case OP_IMAGE_TEXT8:
				if (ptr->c_req_len <= len)
					VXFilterImageText8 (ptr);
				if (ptr->input_temp)
					len = 0;
				else
					len -= ptr->c_req_len;
				break;
			case OP_POLY_TEXT8:
				VXFilterPolyText8(ptr);
				len = 0;
				break;
			case OP_CREATE_GC:
			case OP_CHANGE_GC:
				/* process after reading entire request */
				if(ptr->c_req_len <= len) {
					VXFilterGC(ptr);
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
				} else
					len = 0;
				break;
			case OP_FREE_GC:
				/* process after reading entire request */
				if(ptr->c_req_len <= len) {
					VXFilterFreeGC(ptr);
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
				} else
					len = 0;
				break;
			case OP_COPY_GC:
				/* process after reading entire request */
				if(ptr->c_req_len <= len) {
					VXFilterCopyGC(ptr);
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
				} else
					len = 0;
				break;
			default:
				/* copy data */
				if(ptr->c_req_len <= len) {
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
					if(ptr->c_connected == 0) {
						ptr->c_connected = 1;
					}
				} else {
					VXMemCopyClient(ptr, len);
					len = 0;
				}
				break;
			}
		}
	}
#else
	len = MAX_BUFFER - ptr->so_buff_size;
	if(ptr->ci_buff_size < len)
		len = ptr->ci_buff_size;
	if(len > 0)		/* buffer is full || no data */
		VXMemCopyClient(ptr, len);
#endif
}

void
VXFilterServerInput(Channel *ptr)
{
	int  len;
	unsigned short  seq_num;

#ifndef	_NO_FILTER_
	/* len = min(available space, number of data to be output) */
	len = MAX_BUFFER - ptr->co_buff_size;
	if(ptr->si_buff_size < len)
		len = ptr->si_buff_size;
	if(len == 0)		/* buffer is full || no data */
		return;
	
	while(len > 0) {
		if(ptr->s_req_len == 0) {
			ptr->s_ipos = 0;
			if(ptr->s_connected == 0) {
				if(ptr->si_buffer[0] == 1) {
					if(ptr->si_buff_size >= 24) {
						ptr->res_base = mem_long(ptr->si_buffer, 12);
						ptr->res_mask = mem_long(ptr->si_buffer, 16);
						TRACE2("BASE, MASK = %x, %x\n", ptr->res_base, ptr->res_mask);
					} else
						break;
				}
				if(ptr->si_buff_size >= 8) {
					ptr->s_req_len = mem_short(ptr->si_buffer, 6) * 4;
					ptr->s_req_len += 8;
					ptr->s_connected = 1;
				}
			} else {
				if(ptr->si_buff_size >= 8) {
					seq_num = mem_short(ptr->si_buffer, 2);
					if(seq_num > ADDITIONAL_REQUESTS) {
						seq_num -= ADDITIONAL_REQUESTS;
						mem_set_short(ptr->si_buffer, 2, seq_num);
					}
					if(ptr->si_buffer[0] != 1)
						ptr->s_req_len = 32;
					else if(ptr->si_buff_size >= 8)
						ptr->s_req_len = 32 + mem_long(ptr->si_buffer, 4)*4;

					/*
					if(ptr->si_buffer[0] == OP_KEY_PRESS) {
						printf("key = %c\n", ptr->si_buffer[1]);
					}
					*/
				}
			}
		}
	
		if(ptr->s_req_len > 0) {
			if(ptr->s_req_len <= len) {
				len -= ptr->s_req_len;
				VXMemCopyServer(ptr, ptr->s_req_len);
			} else {
				VXMemCopyServer(ptr, len);
				len = 0;
			}
		} else
			break;
	}
#else
	len = MAX_BUFFER - ptr->co_buff_size;
	if(ptr->si_buff_size < len)
		len = ptr->si_buff_size;
	if(len > 0)		/* buffer is full || no data */
		VXMemCopyServer(ptr, len);
#endif
}
