/*    %W% %G%
================================================================================
  Description :
  Author      : Ryu, Byeong-Soon
================================================================================
*/

#include <stdio.h>
#include "channel.h"

#define GC_MASK_FONT	0x00004000
#define GC_BIT_POS_FONT	14

void
VXSetFontGC(Channel *ptr, long gc, long font)
{
	FontGC  *gpt;

	gpt = ptr->font_gc;
	while(gpt != NULL) {
		if(gpt->gc == gc) {
			gpt->font = font;
			return;
		}
		gpt = gpt->next;
	}
	gpt = (FontGC *) malloc(sizeof(FontGC));
	if(gpt == NULL) {
		perror("VXSetFontGC");
		return;
	}
	gpt->next = ptr->font_gc;
	ptr->font_gc = gpt;
	gpt->gc = gc;
	gpt->font = font;
}

long
VXGetFontGC(Channel *ptr, long gc)
{
	FontGC  *gpt;

	gpt = ptr->font_gc;
	while(gpt != NULL) {
		if(gpt->gc == gc) {
			return gpt->font;
		}
		gpt = gpt->next;
	}
	return 0;
}

void
VXFilterGC(Channel *ptr)
{
	int  i, pos_mask, pos_value;
	long  gc, mask, font;

	pos_mask = ptr->opcode == OP_CREATE_GC ? 12 : 8;
	pos_value = pos_mask + 4;

	gc = mem_long(ptr->ci_buffer, 4);
	mask = mem_long(ptr->ci_buffer, pos_mask);
	if(mask & GC_MASK_FONT) {
		for(i = 0; i < GC_BIT_POS_FONT; i++) {
			if(mask&0x00000001)
				pos_value += 4;
			mask >>= 1;
		}
		fprintf(stderr, "pos_value = %d\n", pos_value);
		font = mem_long(ptr->ci_buffer, pos_value);
		VXSetFontGC(ptr, gc, font);
	}
}

void
VXFilterCopyGC(Channel *ptr)
{
	long  src, dst, mask, font;

	src = mem_long(ptr->ci_buffer, 4);
	dst = mem_long(ptr->ci_buffer, 8);
	mask = mem_long(ptr->ci_buffer, 12);

	if(mask & GC_MASK_FONT) {
		font = VXGetFontGC(ptr, src);
		VXSetFontGC(ptr, dst, font);
	}
}

void
VXFilterFreeGC(Channel *ptr)
{
	FontGC  *gpt, *pre;
	long  gc;

	gc = mem_long(ptr->ci_buffer, 4);
	gpt = ptr->font_gc;
	pre = NULL;
	while(gpt != NULL) {
		if(gpt->gc == gc) {
			if(pre == NULL)
				ptr->font_gc = gpt->next;
			else
				pre->next = gpt->next;
			free(gpt);
			return;
		}
		pre = gpt;
		gpt = gpt->next;
	}
}
